@@ -1,5 +1,6 @@
 use crate::ansi::remove_ansi_escape_codes;
 use crate::commit_line::format_commit_line;
+use crate::hunk_header::HunkHeader;
 use crate::io::ErrorKind;
 use crate::refiner::to_highlighted_tokens;
 use crate::token_collector::{
@@ -15,8 +16,6 @@ use std::thread::{self, JoinHandle};
 use crate::{constants::*, refiner};
 use threadpool::ThreadPool;
 
-const HUNK_HEADER: &str = "\x1b[36m"; // Cyan
-
 lazy_static! {
     static ref STATIC_HEADER_PREFIXES: Vec<(&'static str, &'static str)> = vec![
         ("diff ", FAINT),
@@ -147,12 +146,26 @@ The diff lines blocks will also be enqueued for printing, but the actual diffing
 will happen in background threads.
 */
 pub struct LineCollector {
-    old_text: String,
-    new_text: String,
-    plain_text: String,
-    diff_seen: bool,
-    consumer_thread: Option<JoinHandle<()>>,
+    /// Calculated by HunkHeader::parse(). We'll count this value down as we consume lines.
+    expected_old_lines: usize,
 
+    /// Calculated by HunkHeader::parse(). We'll count this value down as we consume lines.
+    expected_new_lines: usize,
+
+    /// The old text of a diff, if any. Includes `-` lines only.
+    old_text: String,
+
+    /// The new text of a diff, if any. Includes `+` lines only.
+    new_text: String,
+
+    /// Headers and stuff that we just want printed, not part of a diff
+    plain_text: String,
+
+    /// Set to true when we see the first diff line. The second diff line and
+    /// onwards will come with highlighted backgrounds, based on this value.
+    diff_seen: bool,
+
+    consumer_thread: Option<JoinHandle<()>>,
     diffing_threads: ThreadPool,
 
     // FIXME: I'd rather have had a SyncSender of some trait here. That would
@@ -218,6 +231,8 @@ impl LineCollector {
             .unwrap();
 
         return LineCollector {
+            expected_old_lines: 0,
+            expected_new_lines: 0,
             old_text: String::from(""),
             new_text: String::from(""),
             plain_text: String::from(""),
@@ -306,23 +321,25 @@ impl LineCollector {
         ))
     }
 
-    pub fn consume_plusminus_header(&mut self, line: &str) {
+    /// Returns an error message on failure
+    #[must_use]
+    pub fn consume_plusminus_header(&mut self, line: &str) -> Option<&str> {
         if let Some(old_name) = line.strip_prefix("--- ") {
             self.old_text.clear();
             self.old_text.push_str(old_name);
-            return;
+            return None;
         }
 
         if let Some(new_name) = line.strip_prefix("+++ ") {
             if self.old_text.is_empty() {
                 // We got +++ not preceded by ---, WTF?
-                return;
+                return None;
             }
 
             self.new_text.clear();
             self.new_text.push_str(new_name);
         } else {
-            panic!("Got a plusminus header that doesn't start with --- or +++");
+            return Some("Got a plusminus header that doesn't start with --- or +++");
         }
 
         if self.old_text == "/dev/null" {
@@ -338,7 +355,7 @@ impl LineCollector {
             self.consume_plain_linepart("+++ ");
             self.consume_plain_linepart(&new_name);
             self.consume_plain_line(NORMAL);
-            return;
+            return None;
         }
 
         if self.new_text == "/dev/null" {
@@ -355,7 +372,7 @@ impl LineCollector {
             self.consume_plain_linepart("+++ /dev/null");
             self.consume_plain_line(NORMAL);
 
-            return;
+            return None;
         }
 
         let (mut old_tokens, mut new_tokens, _, _) =
@@ -372,72 +389,20 @@ impl LineCollector {
         let new_filename = render(&LINE_STYLE_NEW_FILENAME, new_tokens);
         self.consume_plain_line(&old_filename);
         self.consume_plain_line(&new_filename);
-    }
 
-    fn consume_hunk_header(&mut self, line: &str) {
-        self.consume_plain_linepart(HUNK_HEADER);
-
-        if let Some(second_atat_index) = line.find(" @@ ") {
-            // Highlight the function name
-            self.consume_plain_linepart(FAINT);
-            self.consume_plain_linepart(&line[..(second_atat_index + 4)]);
-            self.consume_plain_linepart(BOLD);
-            self.consume_plain_linepart(&line[(second_atat_index + 4)..]);
-        } else {
-            self.consume_plain_linepart(line);
-        }
-
-        self.consume_plain_line(NORMAL);
+        return None;
     }
 
     /// The line parameter is expected *not* to end in a newline
-    pub fn consume_line(&mut self, line: &mut Vec<u8>) {
+    ///
+    /// Returns an error message on trouble.
+    #[must_use]
+    pub fn consume_line(&mut self, line: &mut Vec<u8>) -> Option<&str> {
         // Strip out incoming ANSI formatting. This enables us to highlight
         // already-colored input.
         remove_ansi_escape_codes(line);
         let line = String::from_utf8_lossy(line);
 
-        if line.starts_with("diff") {
-            self.diff_seen = true;
-        }
-
-        if let Some(fixed_highlight) = get_fixed_highlight(&line) {
-            self.consume_plain_linepart(fixed_highlight);
-            self.consume_plain_linepart(&line);
-            self.consume_plain_line(NORMAL); // consume_plain_line() will add a linefeed to the output
-            return;
-        }
-
-        if line.starts_with("commit") {
-            self.consume_plain_line(&format_commit_line(&line, self.diff_seen));
-            return;
-        }
-
-        if line.starts_with("--- ") || line.starts_with("+++ ") {
-            self.consume_plusminus_header(&line);
-            return;
-        }
-
-        if line.starts_with("@@ ") {
-            self.consume_hunk_header(&line);
-            return;
-        }
-
-        if line.is_empty() {
-            self.consume_plain_line("");
-            return;
-        }
-
-        if line.starts_with('-') {
-            self.consume_old_line(&line);
-            return;
-        }
-
-        if line.starts_with('+') {
-            self.consume_new_line(&line);
-            return;
-        }
-
         if line.starts_with('\\') {
             {
                 // Store the "\ No newline at end of file" string however it is
@@ -458,9 +423,65 @@ impl LineCollector {
             // condition.
             self.consume_no_eof_newline_marker(&line);
 
-            return;
+            return None;
+        }
+
+        if self.expected_old_lines + self.expected_new_lines > 0 {
+            if line.starts_with('-') {
+                self.expected_old_lines -= 1;
+                self.consume_old_line(&line);
+                return None;
+            }
+
+            if line.starts_with('+') {
+                self.expected_new_lines -= 1;
+                self.consume_new_line(&line);
+                return None;
+            }
+
+            if !line.is_empty() && !line.starts_with(' ') {
+                return Some("Non-empty line should have started with ' ', '-' or '+'");
+            }
+
+            self.expected_old_lines -= 1;
+            self.expected_new_lines -= 1;
+            self.consume_plain_line(&line);
+            return None;
+        }
+
+        if let Some(hunk_header) = HunkHeader::parse(&line) {
+            self.expected_new_lines = hunk_header.new_linecount;
+            self.expected_old_lines = hunk_header.old_linecount;
+            self.consume_plain_line(&hunk_header.render());
+            return None;
+        }
+
+        if line.starts_with("diff") {
+            self.diff_seen = true;
+        }
+
+        if let Some(fixed_highlight) = get_fixed_highlight(&line) {
+            self.consume_plain_linepart(fixed_highlight);
+            self.consume_plain_linepart(&line);
+            self.consume_plain_line(NORMAL); // consume_plain_line() will add a linefeed to the output
+            return None;
+        }
+
+        if line.starts_with("commit") {
+            self.consume_plain_line(&format_commit_line(&line, self.diff_seen));
+            return None;
+        }
+
+        if line.starts_with("--- ") || line.starts_with("+++ ") {
+            return self.consume_plusminus_header(&line);
+        }
+
+        if line.is_empty() {
+            self.consume_plain_line("");
+            return None;
         }
 
         self.consume_plain_line(&line);
+        return None;
     }
 }
